import numpy as np
from pymongo import MongoClient

default_name = 'faiss_storage'


def get_mongo_client():
    try:
        result = MongoClient(connectTimeoutMS=1500, socketTimeoutMS=1500)
        ping = result[default_name].command('ping')
        if ping['ok']:
            return result
    except Exception as e:
        print(f'Can not connect to default mongo connection')

class Storage:
    def __init__(self, name=default_name):
        self.mongo = get_mongo_client()
        if not self.mongo:
            raise Exception('Can not connect to mongo!')
        self.db = self.mongo[name]


class AccumulationStorage(Storage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if 'data' in self.db.collection_names(include_system_collections=False):
            self.data_col = self.db.get_collection('data')
        else:
            self.data_col = self.db.create_collection('data')

        if 'meta' in self.db.collection_names(include_system_collections=False):
            self.meta_col = self.db.get_collection('meta')
        else:
            self.meta_col = self.db.create_collection('meta')

        self.counter = (self.meta_col.find_one({'_id': 'counter'}) or {}).get('val', 0)
        if self.counter == 0:
            self.meta_col.insert_one({'_id': 'counter', 'val': 0})

    def accumulate(self, vector: np.ndarray, grid_id: int):
        self.data_col.insert_one({'_id': self.counter, 'vec': vector.dumps(), 'gid': grid_id})
        self.meta_col.update_one({'_id': 'counter'}, {'$inc': {'val': 1}})
        self.counter += 1

    def iterate(self):
        for el in self.data_col.find():
            yield {'id': el['_id'],
                   'vector': np.loads(el['vec']),
                   'grid_id': el['gid']
                   }


class VectorStorage(Storage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.vectors_coll = self.db.get_collection('vectors') or self.db.create_collection('vectors')

    def add_vector_info(self, vector_info):
        self.vectors_coll.insert_one(vector_info)

    def get_by_ids(self, ids):
        return [r for r in self.vectors_coll.find({'_id': {'$in': ids}})]


def get_storages():
    return AccumulationStorage(), VectorStorage()
