import json
import random

import numpy as np
import requests
import tornado

from base import controller
from common.faiss_index import faiss_index
from settings import RESNET_URL, FAISS_DISTANCE_DELTA, INDEX_DUMP_TIMEOUT


def dump_index():
    faiss_index.save()


tornado.ioloop.PeriodicCallback(dump_index, INDEX_DUMP_TIMEOUT * 1000 * 60).start()


class FaissServiceBaseController(controller.Controller):

    def process(self, vector):
        pass

    def post(self):
        try:
            check_args = self.args_required([
                "lat",
                "lon",
                "transform",
                "accuracy"
            ])
            if check_args:
                self.args['lat'] = float(self.args['lat'])
                self.args['lon'] = float(self.args['lon'])
                self.args['accuracy'] = float(self.args['accuracy'])
                self.args['transform'] = [
                    float(x) for x in self.args['transform'].strip('(\[|\])').split(',')]
                file_data = self.request.files['img'][0]
                self.args['grid_id'] = self.args.get('grid_id', None)
            else:
                raise Exception('Not enough arguments')
        except KeyError as e:
            return self.error_message(400, f'need: {e}')
        except Exception as e:
            return self.error_message(400, f'exc at parse args: {e}')

        result = {}
        rs_params = {'files': {'img': file_data['body']}, 'data': {
            'lat': self.args['lat'], 'lon': self.args['lon'], 'accuracy': self.args['accuracy']}, 'timeout': 1.5}
        rs_resp = requests.post(RESNET_URL, **rs_params)
        if rs_resp.status_code != 200:
            return self.error_message(rs_resp.status_code, f'Error at RESNET side: {rs_resp.content}')
        try:
            data = json.loads(rs_resp.content.decode())
            rs_vector = data.get('data').get('vector')
            rs_vector = np.array([rs_vector], dtype='float32')
            result = self.process(rs_vector)
        except Exception as e:
            print(f'error at parse result: {e}')
            return self.error_message(500, f'error: {e}')
        return self.write_json({'result': result or []})


class AddController(FaissServiceBaseController):

    def post(self):
        """
        ---
        tags:
        - Add image to faiss index
        summary: Use resnet service for add image to index
        description: Use resnet service for add image to index
        consumes:
        - "multipart/form-data"
        produces:
        - "application/json"
        parameters:
        - name: "lat"
          in: "formData"
          description: "Latitude"
          required: true
          type: "number"
          format: "float"
        - name: "lon"
          in: "formData"
          description: "Longitude"
          required: true
          type: "number"
          format: "float"
        - name: "accuracy"
          in: "formData"
          description: "GPS Accuracy"
          required: true
          type: "number"
          format: "float"
        - name: "transform"
          in: "formData"
          description: "Transform of user"
          required: true
          type: "array"
          items:
            type: "number"
            format: "float"
        - name: "grid_id"
          in: "formData"
          description: "Grid uuid"
          required: true
          type: "string"
        - name: "img"
          description: "File of image"
          required: true
          type: file

        responses:
            200:
                description: success
                schema:
                    type: "object"
                    properties:
                        result:
                            type: "boolean"
            400:
                description: if some parameters are not present
            500:
                description: if something went wrong
        """
        return super(AddController, self).post()

    def process(self, rs_vector):
        if self.args['grid_id']:
            d, i = faiss_index.rate(rs_vector, 1)
            distance = float(d[0][0])
            if distance > FAISS_DISTANCE_DELTA or distance < 0:
                self.args["_id"] = random.randint(0, 1 << 63)
                self.vectors_store.add_vector_info(self.args)
                faiss_index.add(rs_vector, np.array([self.args["_id"]]))
                faiss_index.save()

                self.accumulate_store.accumulate(rs_vector, self.args['grid_id'])

                return True
        else:
            raise Exception('need grid_id')


class RateController(FaissServiceBaseController):

    def post(self):
        """
        ---
        tags:
        - Rate image
        summary: Use resnet service for rate image
        description: Use resnet service for rate image
        consumes:
        - "multipart/form-data"
        produces:
        - "application/json"
        parameters:
        - name: "lat"
          in: "formData"
          description: "Latitude"
          required: true
          type: "number"
          format: "float"
        - name: "lon"
          in: "formData"
          description: "Longitude"
          required: true
          type: "number"
          format: "float"
        - name: "transform"
          in: "formData"
          description: "Transform of user"
          required: true
          type: "array"
          items:
            type: "number"
            format: "float"
        - name: "img"
          description: "File of image"
          required: true
          type: file
        - name: "accuracy"
          in: "formData"
          description: "GPS Accuracy"
          required: true
          type: "number"
          format: "float"

        responses:
            200:
                description: success
                schema:
                    type: "object"
                    properties:
                        result:
                            type: "array"
                            items:
                                type: object
                                properties:
                                    distance:
                                        type: "number"
                                    _id:
                                        type: "number"
            400:
                description: if some parameters are not present
                schema:
                    type: object
                    properties:
                        message:
                            type: string
            500:
                description: if something went wrong
        """
        return super(RateController, self).post()

    def process(self, rs_vector):
        dst, idx = faiss_index.rate(rs_vector, 30)
        result_map = {}
        for i in range(len(idx[0])):
            result_map[idx[0][i]] = dst[0][i]
        db_results = self.vectors_store.get_by_ids(idx[0].tolist())

        result = {}
        for r in db_results:
            g_id = r['grid_id']
            dst = result_map[r['_id']]
            r['distance'] = dst
            result[g_id] = r if g_id not in result else r if dst < result[g_id]['distance'] else result[g_id]

        return list(result.values())

# curl --verbose --request POST --header "Content-Type:multipart/form-data" --form grid_id=qwerty --form lat=10 --form lon=8 --form transform=[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0] --form img=@/Users/vladislavgaidukov/avatar.jpg  http://localhost:8888/add
