import json
import mimetypes
import random
from functools import partial
from uuid import uuid4

import numpy as np
import requests
from mock import patch
from tornado import gen
from tornado.ioloop import IOLoop
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application

from common.faiss_index import FaissIndex
from server import FAISSApplication
from settings import RESNET_URL


class FakeAccStorage:
    def __init__(self, **kwargs):
        self.data = []

    def accumulate(self, vector: np.ndarray, grid_id: int):
        self.data.append((vector, grid_id))

    def iterate(self):
        for i, (v, gid) in enumerate(self.data):
            yield {'id': i, 'grid_id': gid, 'vector': v}


class FakeVecStorage:
    def __init__(self, **kwargs):
        self.data = {}

    def add_vector_info(self, vinfo):
        self.data[vinfo['_id']] = vinfo

    def get_by_ids(self, ids):
        inter = set(self.data.keys()).intersection(set(ids))
        return [v for k, v in self.data.items() if k in inter]


fake_acc_store = FakeAccStorage()
fake_vec_store = FakeVecStorage()


def get_fake_storages():
    return fake_acc_store, fake_vec_store


def get_vector(filename):
    with open(filename, "rb") as f:
        lat, lon = random_coord()
        rs_params = {'files': {'img': f}, 'data': {'lat': lat, 'lon': lon, 'accuracy': 10}, 'timeout': 1.5}
        rs_resp = requests.post(RESNET_URL, **rs_params)
        data = rs_resp.json()
        rs_vector = data.get('data').get('vector')
        return rs_vector


class AppTestSetup(AsyncHTTPTestCase):

    def get_app(self, ) -> Application:
        return FAISSApplication()

    def get_new_ioloop(self) -> IOLoop:
        return IOLoop.current()


@gen.coroutine
def multipart_producer(boundary, filename, grid_id, write):
    boundary_bytes = boundary.encode()

    filename_bytes = filename.encode()
    mtype = mimetypes.guess_type(filename)[0] or "application/octet-stream"
    buf = (
            (b"--%s\r\n" % boundary_bytes)
            + (
                    b'Content-Disposition: form-data; name="img"; filename="%s"\r\n'
                    % (filename_bytes)
            )
            + (b"Content-Type: %s\r\n" % mtype.encode())
            + b"\r\n"
    )
    yield write(buf)
    with open(filename, "rb") as f:
        while True:
            chunk = f.read(16 * 1024)
            if not chunk:
                break
            yield write(chunk)

    yield write(b"\r\n")

    buf = (
            b'--%s\r\n' % boundary_bytes
            + b'Content-Disposition: form-data; name="lat"\r\n'
            + b'\r\n3.123313'
            + b'\r\n--%s\r\n' % boundary_bytes
            + b'Content-Disposition: form-data; name="lon"\r\n'
            + b'\r\n2.22121'
            + b'\r\n--%s\r\n' % boundary_bytes
            + b'Content-Disposition: form-data; name="grid_id"\r\n'
            + b'\r\nqwerty'
            + b'\r\n--%s\r\n' % boundary_bytes
            + b'Content-Disposition: form-data; name="accuracy"\r\n'
            + b'\r\n10'
            + b'\r\n--%s\r\n' % boundary_bytes
            + b'Content-Disposition: form-data; name="transform"\r\n'
            + b'\r\n[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16]'
    )
    if grid_id:
        buf += (
                b'\r\n--%s\r\n' % boundary_bytes
                + b'Content-Disposition: form-data; name="grid_id"\r\n'
                + b'\r\n%s' % str(grid_id).encode()
        )
    yield write(buf)

    yield write(b"\r\n")

    yield write(b"--%s--\r\n" % (boundary_bytes,))


from random import uniform


def random_coord():
    return uniform(-90, 90), uniform(-180, 180)


class MainControllerTestCase(AppTestSetup):
    def setUp(self) -> None:
        super().setUp()
        self.index = FaissIndex('index_test')

    @patch('base.controller.get_storages', side_effect=get_fake_storages)
    def test_main_controller_add_to_index(self, fn):
        boundary = uuid4().hex
        headers = {"Content-Type": "multipart/form-data; boundary=%s" % boundary}
        producer = partial(multipart_producer, boundary, './tests/test.jpg', None)
        response = self.fetch('/add', method="POST", body_producer=producer, headers=headers)
        data = json.loads(response.body.decode('utf-8'))
        print(data)
        self.assertIsNotNone(data['data'].get('result'))

    def test_add_and_rate_vector(self):
        data = get_vector('./tests/test.jpg')
        vid = random.randint(0, 1 << 63)
        vid = np.array([vid])
        self.index.add(np.array([data], dtype="float32"), vid)
        d, i = self.index.rate(np.array([data], dtype="float32"), 1)
        self.assertEqual(i[0], vid[0])

    @patch('base.controller.get_storages', side_effect=get_fake_storages)
    def test_main_controller_rate_index(self, fn):
        boundary = uuid4().hex
        headers = {"Content-Type": "multipart/form-data; boundary=%s" % boundary}
        producer = partial(multipart_producer, boundary, './tests/test.jpg', 100500600)
        response = self.fetch('/add', method="POST", body_producer=producer, headers=headers)
        data = json.loads(response.body.decode('utf-8'))
        self.assertIsNotNone(data['data'].get('result'))
# curl --verbose --request POST --header "Content-Type:multipart/form-data" --form lat=10 --form lon=10 --form transform=[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0] --form img=@/Users/vladislavgaidukov/Downloads/2011_09_29/2011_09_29_drive_0071_sync/image_00/data/0000000448.png   http://localhost:8888/
